unit Macros;

{This unit contains the recursive descent parser/}
{interpreter for Image's Pascal-like macro language.}

{References:}
{  "Pascal User Manual and Report", Kathleen Jensen and Niklaus Wirth, Springer-Verlag}
{  "Building Your Own C Interpreter", Dr. Dobb's Journal, August 1989}


interface

	uses
		QuickDraw, PrintTraps, globals, Utilities, Graphics, edit, Analysis, Camera, file1, file2, functions;


	procedure LoadMacros (fname: str255; RefNum: integer);
	procedure RunMacro (nMacro: integer);
	procedure RunKeyMacro (ch: char; KeyCode: integer);


implementation


	const
		DivideByZero = 'Divide by zero';
		DoExpected = '"do" expected';
		StackOverflow = 'Stack overflow';
		RightParenExpected = '")" expected';
		MaxArgs = 25;

	var
		EndMacros, pcStart, TopOfStack, CurrentH, CurrentV, nSymbols, TokenLoc: integer;
		MacroStart: array[1..MaxMacros] of integer;
		MacroKey: packed array[1..MaxMacros] of char;
		pc: integer; {"program counter" used by macro interpreter}
		SavePC: integer;
		Token, SaveToken: TokenType;
		TokenStr: string[100];
		TokenSymbol, ProcName: SymbolType;
		TokenValue: extended;
		TokenStackLoc: integer;
		Command: CommandType;


	function GetExpression: extended;
	forward;
	procedure DoStatement;
	forward;
	procedure DoFor;
	forward;



	function LineNumber: integer;
		var
			i, n: integer;
	begin
		n := 1;
		with MacrosP^ do
			for i := pcStart to pc do
				if macros[i] = cr then
					n := n + 1;
		LineNumber := n;
	end;


	function PreScan: boolean;
   {Convert to lowercase and remove comments to speed up parsing.}
		var
			inString, inComment: boolean;
			c: char;
			i, StartComment: integer;
	begin
		with MacrosP^ do begin
				PreScan := false;
				inString := false;
				inComment := false;
				pcStart := 0;
				for i := 0 to EndMacros do begin
						c := macros[i];
						if inString and (c = cr) then begin
								pc := i;
								PutMessage(concat('The quoted string in line ', long2str(LineNumber), ' is not terminated.'));
								exit(PreScan);
							end;
						if (not InString) and (c = '{') then begin
								InComment := true;
								StartComment := i;
							end;
						if inComment then begin
								if (c = '{') and (i <> StartComment) then begin
										PutMessage('Comments cannot be nested.');
										exit(PreScan);
									end;
								if c = '}' then
									inComment := false;
								macros[i] := ' ';
							end
						else begin
								if c = '''' then
									inString := not inString;
								if (c >= 'A') and (c <= 'Z') and not inString then
									macros[i] := chr(ord(c) + 32);
							end;
					end;
				if inComment then begin
						pc := StartComment;
						PutMessage(concat('The comment starting in line ', long2str(LineNumber), ' is not terminated.'))
					end
				else
					PreScan := true;
			end; {with}
	end;


	function match (str: str255): boolean;
		var
			i, loc: integer;
	begin
		loc := pc - 1;
		match := false;
		with MacrosP^ do
			for i := 1 to length(str) do
				if macros[loc + i] <> str[i] then
					exit(match);
		match := true;
	end;


	procedure TrimString (var str: str255);
	begin
		if length(str) > 0 then begin
				while (length(str) > 1) and (str[1] = ' ') do
					delete(str, 1, 1);
				while (length(str) > 1) and ((str[length(str)] = ' ') or (str[length(str)] = ';')) do
					delete(str, length(str), 1);
			end;
	end;


	procedure MacroError (str: str255);
		var
			cLine, str2: str255;
			i, count: integer;
	begin
		with MacrosP^ do begin
				if token = DoneT then
					exit(MacroError);
				i := pc - 1;
				while (i > 0) and (macros[i] <> cr) do {Find start of line}
					i := i - 1;
				count := 0;
				cLine := '';
				repeat
					i := i + 1;
					count := count + 1;
					if macros[i] <> cr then
						cLine := concat(cLine, macros[i]);
				until (i >= EndMacros) or (macros[i] = cr) or (count > 60);
				TrimString(cLine);
				if ProcName <> '        ' then begin {Are we in a procedure?}
						str2 := ProcName;
						TrimString(str2);
						str2 := concat(' of procedure ', str2)
					end
				else
					str2 := '';
				PutMessage(concat(str, ' in line ', long2str(LineNumber), str2, '.', cr, cr, '"', cLine, '"'));
				Token := DoneT;
			end; {with}
	end;


	procedure LookupIdentifier;
  {Binary search routine from "Data Structures with Abstract}
  {Data Types and Pascal" by Stubbs and Webre.}
		var
			low, high, mid: integer;
	begin
		with MacrosP^ do begin
				low := 1;
				high := nSymbols;
				while low < high do begin
						mid := (low + high + 1) div 2;
						if TokenSymbol < SymbolTable[mid].symbol then
							high := mid - 1
						else
							low := mid;
					end;
				with SymbolTable[high] do
					if (high <> 0) and (TokenSymbol = symbol) then begin
							token := tType;
							Command := cType;
							TokenLoc := loc;
						end
					else
						token := UnknownIdentifier;
			end;
	end;


	procedure LookupVariable;
		var
			VarFound: boolean;
			i: integer;
	begin
		with MacrosP^ do begin
				VarFound := false;
				i := TopOfStack + 1;
				repeat
					i := i - 1;
					VarFound := TokenSymbol = Stack[i].symbol
				until VarFound or (i = 1);
				if VarFound then
					with stack[i] do begin
							token := Variable;
							TokenValue := value;
							TokenStackLoc := i;
						end;
			end; {with}
	end;


	procedure GetToken;
		var
			c: char;
			SymbolLength: integer;
	begin
		with MacrosP^ do begin
				if token = DoneT then
					exit(GetToken);
				SavePC := PC;
				SaveToken := token;
				while not (macros[pc] in ['a'..'z', '0'..'9', '(', ')', ',', '''', '+', '-', '*', '/', ':', ';', '=', '.', '>', '<']) do begin  {skip white space}
						pc := pc + 1;
						if pc > EndMacros then begin
								Token := DoneT;
								exit(GetToken);
							end;
					end;
				c := macros[pc];
				case c of
					'a'..'z': 
						begin
							TokenSymbol := '        ';
							SymbolLength := 0;
							while macros[pc] in ['a'..'z', '0'..'9'] do begin
									SymbolLength := SymbolLength + 1;
									if SymbolLength <= 8 then
										TokenSymbol[SymbolLength] := macros[pc];
									pc := pc + 1;
									if pc > EndMacros then begin
											Token := DoneT;
											exit(GetToken);
										end;
								end;
							Token := identifier;
							LookupIdentifier;
							if (token = UnknownIdentifier) and (TopOfStack > 0) then
								LookupVariable;
							exit(GetToken);
						end;
					'0'..'9', '.': 
						begin
							TokenStr := '';
							while macros[pc] in ['0'..'9', '.'] do begin
									TokenStr := concat(TokenStr, c);
									pc := pc + 1;
									c := macros[pc];
									if pc > EndMacros then begin
											Token := DoneT;
											exit(GetToken);
										end;
								end;
							Token := NumericLiteral;
							exit(GetToken);
						end;
					'(': 
						begin
							Token := LeftParen;
							pc := pc + 1;
						end;
					')': 
						begin
							Token := RightParen;
							pc := pc + 1;
						end;
					',': 
						begin
							Token := comma;
							pc := pc + 1;
						end;
					':': 
						if macros[pc + 1] = '=' then begin
								Token := AssignOp;
								pc := pc + 2;
							end
						else begin
								Token := colon;
								pc := pc + 1;
							end;
					';': 
						begin
							Token := SemiColon;
							pc := pc + 1;
						end;
					'+': 
						begin
							Token := PlusOp;
							pc := pc + 1;
						end;
					'-': 
						begin
							Token := MinusOp;
							pc := pc + 1;
						end;
					'*': 
						begin
							Token := MulOp;
							pc := pc + 1;
						end;
					'/': 
						begin
							Token := DivOp;
							pc := pc + 1;
						end;
					'''': 
						begin
							TokenStr := '';
							pc := pc + 1;
							while macros[pc] <> '''' do begin
									TokenStr := concat(TokenStr, macros[pc]);
									pc := pc + 1;
									if pc > EndMacros then begin
											Token := DoneT;
											exit(GetToken);
										end;
								end;
							pc := pc + 1;
							Token := stringT;
						end;
					'=': 
						begin
							Token := eqOp;
							pc := pc + 1;
						end;
					'<': 
						begin
							pc := pc + 1;
							if macros[pc] = '>' then begin
									token := neOp;
									pc := pc + 1;
								end
							else if macros[pc] = '=' then begin
									token := leOp;
									pc := pc + 1;
								end
							else
								token := ltOp;
						end;
					'>': 
						begin
							pc := pc + 1;
							if macros[pc] = '=' then begin
									token := geOp;
									pc := pc + 1;
								end
							else
								token := gtOp;
						end;
					otherwise
						begin
							token := NullT;
							beep;
						end;
				end; {case}
			end; {with}
	end;


	procedure AddMenuItem;
		var
			i, fkey: integer;
			str: str255;
			c, key: char;
	begin
		with MacrosP^ do begin
				pc := pc + 4;
				repeat
					pc := pc + 1;
					c := macros[pc];
				until (c = '''') or (c = cr) or (pc = EndMacros);
				if (c = cr) or (pc = EndMacros) then begin
						MacroError('Menu entry for macro not found.');
						exit(AddMenuItem);
					end;
				pc := pc + 1;
				str := '';
				repeat
					str := concat(str, macros[pc]);
					pc := pc + 1;
				until (macros[pc] = '''') or (pc = EndMacros);
				if pc = EndMacros then
					PutMessage('Ending quote('') missing.')
				else
					pc := pc + 1;
				AppendMenu(SpecialMenuH, str);
				if nMacros < MaxMacros then
					nMacros := nMacros + 1
				else
					beep;
				if macros[pc] = ';' then
					pc := pc + 1;
				MacroStart[nMacros] := pc;
				i := pos('[', str);
				if i > 0 then begin {Assign a key to macro?}
						i := i + 1;
						key := str[i];
						if (key >= 'A') and (key <= 'Z') then
							key := chr(ord(key) + 32);
						MacroKey[nMacros] := key;
						if (key = 'f') and (str[i + 1] in ['1'..'9']) then begin {Function Key?}
								fkey := ord(str[i + 1]) - ord('0');
								if str[i + 2] in ['0'..'5'] then
									fkey := fkey * 10 + ord(str[i + 2]) - ord('0');
								if (fkey >= 1) and (fkey <= 15) then
									MacroKey[nMacros] := chr(ord('A') + fkey - 1);
							end; {Function key?}
					end;
			end; {with}
	end;


	procedure AddProcedure;
	begin
		pc := pc + 9;
		GetToken;
		if token <> UnknownIdentifier then begin
				MacroError('Procedure name missing or previously defined');
				exit(AddProcedure);
			end;
		if nSymBols >= MaxSymbols then begin
				MacroError('Too many procedures');
				exit(AddProcedure);
			end;
		nSymbols := nSymbols + 1;
		with MacrosP^, MacrosP^.SymbolTable[nSymbols] do begin
				symbol := TokenSymbol;
				tType := procedureT;
				cType := NullC;
				if macros[pc] = ';' then
					pc := pc + 1;
				loc := pc;
			end;
	end;


	procedure PreScan2;
		var
			i: integer;
			inString: boolean;
			c: char;
	begin
		Token := NullT;
		with MacrosP^ do begin
				if nMacros > 0 then
					for i := 1 to nMacros do begin
							DelMenuItem(SpecialMenuH, FirstMacroItem);
							MacroKey[i] := chr(0);
						end;
				nMacros := 0;
				pc := 0;
				inString := false;
				while pc <= (EndMacros - 10) do begin
						c := macros[pc];
						if c = '''' then
							inString := not inString;
						if not InString then
							case c of
								'm': 
									if match('macro') then begin
											AddMenuItem;
											if token = DoneT then
												exit(PreScan2);
										end;
								'p': 
									if match('procedure') then begin
											AddProcedure;
											if token = DoneT then
												exit(PreScan2);
										end;
								otherwise
							end;
						pc := pc + 1;
					end; {while}
				if nMacros = 0 then
					PutMessage('No macros found in this file.');
			end; {with}
	end;


	procedure SortSymbolTable;
  {Selection sort routine from "Algorithms" by Robert Sedgewick.}
		var
			i, j, min: integer;
			t: SymTabRec;
	begin
		with MacrosP^ do
			for i := 1 to nSymbols do begin
					min := i;
					for j := i + 1 to nSymbols do
						if SymbolTable[j].symbol < SymbolTable[min].symbol then
							min := j;
					t := SymbolTable[min];
					SymbolTable[min] := SymbolTable[i];
					SymbolTable[i] := t;
				end;
	end;


	procedure InitSymbolTable;
		var
			i: integer;
	begin
		with MacrosP^ do begin
				nSymbols := 95;
				for i := 1 to MaxSymbols do
					with SymbolTable[i] do begin
							tType := CommandT;
							cType := NullC;
							loc := 0;
						end;

				with SymbolTable[1] do begin
						symbol := 'begin   ';
						tType := BeginT;
					end;
				with SymbolTable[2] do begin
						symbol := 'convolve';
						cType := ConvolveC;
					end;
				with SymbolTable[3] do begin
						symbol := 'fill    ';
						cType := FillC;
					end;
				with SymbolTable[4] do begin
						symbol := 'selectal';
						cType := SelectC;
					end;
				with SymbolTable[5] do begin
						symbol := 'invert  ';
						cType := InvertC;
					end;
				with SymbolTable[6] do begin
						symbol := 'killroi ';
						cType := KillC;
					end;
				with SymbolTable[7] do begin
						symbol := 'marksele';
						cType := MarkC;
					end;
				with SymbolTable[8] do begin
						symbol := 'drawboun';
						cType := FrameC;
					end;
				with SymbolTable[9] do begin
						symbol := 'measure ';
						cType := MeasureC;
					end;
				with SymbolTable[10] do begin
						symbol := 'fliphori';
						cType := FliphC;
					end;
				with SymbolTable[11] do begin
						symbol := 'analyzep';
						cType := AnalyzeC;
					end;
				with SymbolTable[12] do begin
						symbol := 'nextwind';
						cType := NextC;
					end;
				with SymbolTable[13] do begin
						symbol := 'paste   ';
						cType := PasteC;
					end;
				with SymbolTable[14] do begin
						symbol := 'clear   ';
						cType := ClearC;
					end;
				with SymbolTable[15] do begin
						symbol := 'copy    ';
						cType := CopyC;
					end;
				with SymbolTable[16] do begin
						symbol := 'restorer';
						cType := RestoreC;
					end;
				with SymbolTable[17] do begin
						symbol := 'threshol';
						cType := ThresholdC;
					end;
				with SymbolTable[18] do begin
						symbol := 'resetgra';
						cType := ResetgmC;
					end;
				with SymbolTable[19] do begin
						symbol := 'makebina';
						cType := MakeBinC;
					end;
				with SymbolTable[20] do begin
						symbol := 'dither  ';
						cType := DitherC;
					end;
				with SymbolTable[21] do begin
						symbol := 'smooth  ';
						cType := SmoothC;
					end;
				with SymbolTable[22] do begin
						symbol := 'smoothmo';
						cType := SmoothMoreC;
					end;
				with SymbolTable[23] do begin
						symbol := 'sharpen ';
						cType := SharpenC;
					end;
				with SymbolTable[24] do begin
						symbol := 'sharpenm';
						cType := SharpenMoreC;
					end;
				with SymbolTable[25] do begin
						symbol := 'showdens';
						cType := SliceC;
					end;
				with SymbolTable[26] do begin
						symbol := 'hidedens';
						cType := NoSliceC;
					end;
				with SymbolTable[27] do begin
						symbol := 'redirect';
						cType := RedirectC;
					end;
				with SymbolTable[28] do begin
						symbol := 'noredire';
						cType := NoRedirectC;
					end;
				with SymbolTable[29] do begin
						symbol := 'wait    ';
						cType := WaitC;
					end;
				with SymbolTable[30] do begin
						symbol := 'resetcou';
						cType := ResetmC;
					end;
				with SymbolTable[31] do begin
						symbol := 'setdensi';
						cType := SetSliceC;
					end;
				with SymbolTable[32] do begin
						symbol := 'undo    ';
						cType := UndoC;
					end;
				with SymbolTable[33] do begin
						symbol := 'setforeg';
						cType := SetForeC;
					end;
				with SymbolTable[34] do begin
						symbol := 'setbackg';
						cType := SetBackC;
					end;
				with SymbolTable[35] do begin
						symbol := 'scalesel';
						cType := ScaleC;
					end;
				with SymbolTable[36] do begin
						symbol := 'rotateri';
						cType := RotateRC;
					end;
				with SymbolTable[37] do begin
						symbol := 'rotatele';
						cType := RotateLC;
					end;
				with SymbolTable[38] do begin
						symbol := 'traceedg';
						cType := TraceC;
					end;
				with SymbolTable[39] do begin
						symbol := 'detected';
						cType := DetectC;
					end;
				with SymbolTable[40] do begin
						symbol := 'showhist';
						cType := HistoC;
					end;
				with SymbolTable[41] do begin
						symbol := 'enhancec';
						cType := EnhanceC;
					end;
				with SymbolTable[42] do begin
						symbol := 'equalize';
						cType := EqualizeC;
					end;
				with SymbolTable[43] do begin
						symbol := 'erode   ';
						cType := ErodeC;
					end;
				with SymbolTable[44] do begin
						symbol := 'dilate  ';
						cType := DilateC;
					end;
				with SymbolTable[45] do begin
						symbol := 'outline ';
						cType := OutlineC;
					end;
				with SymbolTable[46] do begin
						symbol := 'skeleton';
						cType := ThinC;
					end;
				with SymbolTable[47] do begin
						symbol := 'addconst';
						cType := AddConstantC;
					end;
				with SymbolTable[48] do begin
						symbol := 'revertto';
						cType := RevertC;
					end;
				with SymbolTable[49] do begin
						symbol := 'shadow  ';
						cType := ShadowC;
					end;
				with SymbolTable[50] do begin
						symbol := 'flipvert';
						cType := FlipvC;
					end;
				with SymbolTable[51] do begin
						symbol := 'end     ';
						tType := EndT;
					end;
				with SymbolTable[52] do begin
						symbol := 'macro   ';
						tType := DoneT;
					end;
				with SymbolTable[53] do begin
						symbol := 'var     ';
						tType := VarT;
					end;
				with SymbolTable[54] do begin
						symbol := 'for     ';
						tType := ForT;
					end;
				with SymbolTable[55] do begin
						symbol := 'integer ';
						tType := IntegerT;
					end;
				with SymbolTable[56] do begin
						symbol := 'to      ';
						tType := ToT;
					end;
				with SymbolTable[57] do begin
						symbol := 'do      ';
						tType := DoT;
					end;
				with SymbolTable[58] do begin
						symbol := 'beep    ';
						cType := BeepC;
					end;
				with SymbolTable[59] do begin
						symbol := 'capture ';
						cType := CaptureC;
					end;
				with SymbolTable[60] do begin
						symbol := 'makeroi ';
						cType := MakeC;
					end;
				with SymbolTable[61] do begin
						symbol := 'moveroi ';
						cType := MoveC;
					end;
				with SymbolTable[62] do begin
						symbol := 'insetroi';
						cType := InsetC;
					end;
				with SymbolTable[63] do begin
						symbol := 'moveto  ';
						cType := MoveToC;
					end;
				with SymbolTable[64] do begin
						symbol := 'drawtext';
						cType := DrawTextC;
					end;
				with SymbolTable[65] do begin
						symbol := 'setfont ';
						cType := SetFontC;
					end;
				with SymbolTable[66] do begin
						symbol := 'setfonts';
						cType := SetFontSizeC;
					end;
				with SymbolTable[67] do begin
						symbol := 'settext ';
						cType := SetTextC;
					end;
				with SymbolTable[68] do begin
						symbol := 'drawnumb';
						cType := DrawNumC;
					end;
				with SymbolTable[69] do begin
						symbol := 'or      ';
						tType := orOp;
					end;
				with SymbolTable[70] do begin
						symbol := 'and     ';
						tType := andOp;
					end;
				with SymbolTable[71] do begin
						symbol := 'div     ';
						tType := IntDivOp;
					end;
				with SymbolTable[72] do begin
						symbol := 'mod     ';
						tType := modOp;
					end;
				with SymbolTable[73] do begin
						symbol := 'getpicsi';
						cType := GetPicSizeC;
					end;
				with SymbolTable[74] do begin
						symbol := 'putmessa';
						cType := PutMsgC;
					end;
				with SymbolTable[75] do begin
						symbol := 'exit    ';
						cType := ExitC;
					end;
				with SymbolTable[76] do begin
						symbol := 'if      ';
						tType := ifT;
					end;
				with SymbolTable[77] do begin
						symbol := 'then    ';
						tType := thenT;
					end;
				with SymbolTable[78] do begin
						symbol := 'else    ';
						tType := elseT;
					end;
				with SymbolTable[79] do begin
						symbol := 'while   ';
						tType := whileT;
					end;
				with SymbolTable[80] do begin
						symbol := 'repeat  ';
						tType := repeatT;
					end;
				with SymbolTable[81] do begin
						symbol := 'until   ';
						tType := untilT;
					end;
				with SymbolTable[82] do begin
						symbol := 'not     ';
						tType := NotOp;
					end;
				with SymbolTable[83] do begin
						symbol := 'getroi  ';
						cType := GetRoiC;
					end;
				with SymbolTable[84] do begin
						symbol := 'real    ';
						tType := RealT;
					end;
				with SymbolTable[85] do begin
						symbol := 'trunc   ';
						tType := FunctionT;
						cType := TruncC;
					end;
				with SymbolTable[86] do begin
						symbol := 'round   ';
						tType := FunctionT;
						cType := RoundC;
					end;
				with SymbolTable[87] do begin
						symbol := 'getnumbe';
						tType := FunctionT;
						cType := GetNumC;
					end;
				with SymbolTable[88] do begin
						symbol := 'nop     ';
						cType := NopC;
					end;
				with SymbolTable[89] do begin
						symbol := 'random  ';
						tType := FunctionT;
						cType := RandomC;
					end;
				with SymbolTable[90] do begin
						symbol := 'makeoval';
						cType := MakeOvalC;
					end;
				with SymbolTable[91] do begin
						symbol := 'drawscal';
						cType := DrawScaleC;
					end;
				with SymbolTable[92] do begin
						symbol := 'setpalet';
						cType := SetPaletteC;
					end;
				with SymbolTable[93] do begin
						symbol := 'makeneww';
						cType := MakeNewC;
					end;
				with SymbolTable[94] do begin
						symbol := 'getpixel';
						tType := FunctionT;
						cType := GetPixelC;
					end;
				with SymbolTable[95] do begin
						symbol := 'button  ';
						tType := FunctionT;
						cType := ButtonC;
					end;
			end; {with}
	end;


	procedure LoadMacros (fname: str255; RefNum: integer);
		var
			err: OSErr;
			FileSize: LongInt;
			f, SaveNSymbols: integer;
	begin
		err := FSOpen(fname, RefNum, f);
		err := GetEOF(f, FileSize);
		if FileSize > MaxMacroSize then begin
				PutMessage('Macro file is too large.');
				exit(LoadMacros);
			end;
		err := SetFPos(f, fsFromStart, 0);
		err := fsRead(f, FileSize, @MacrosP^.macros);
		EndMacros := FileSize - 1;
		err := fsclose(f);
		ShowWatch;
		if not PreScan then
			exit(LoadMacros);
		InitSymbolTable;
		SortSymbolTable;
		SaveNSymbols := nSymbols;
		PreScan2;
		if nSymbols > SaveNSymbols then
			SortSymbolTable;
		TopOfStack := 0;
		CurrentH := 50;
		CurrentV := 50;
	end;


	procedure GetLeftParen;
	begin
		GetToken;
		if token <> LeftParen then
			MacroError('"(" expected');
	end;


	procedure GetRightParen;
	begin
		GetToken;
		if token <> RightParen then
			MacroError(RightParenExpected);
	end;


	procedure GetComma;
	begin
		GetToken;
		if token <> comma then
			MacroError('"," expected');
	end;


	function GetString: str255;
	begin
		GetToken;
		if token = stringT then
			GetString := TokenStr
		else begin
				MacroError('String expected');
				GetString := '';
			end;
	end;


	procedure DoConvolve;
		var
			err, saveerr: OSErr;
			f, vnum: integer;
			FileFound: boolean;
			fname, vname: str255;
	begin
		GetLeftParen;
		fname := GetString;
		GetRightParen;
		err := fsopen(fname, KernelsRefNum, f);
		FileFound := err = NoErr;
		saveerr := err;
		err := GetVol(@vname, vnum);
{ShowMessage(concat('refnum=', long2str(KernelsRefNum), cr, 'err=', long2str(saveerr), cr, 'name=', fname, cr, 'vname=', vname, cr, 'vnum=', long2str(vnum)));}
{wait(240);}
		err := fsclose(f);
		if FileFound then
			convolve(fname, KernelsRefNum)
		else
			convolve('', 0);
	end;


	procedure CheckBoolean (b: extended);
	begin
		if (b <> ord(true)) and (b <> ord(false)) then
			MacroError('Boolean expression expected');
	end;


	procedure PutTokenBack;
	begin
		if token <> DoneT then begin
				pc := SavePC;
				token := SaveToken;
			end;
	end;


	function DoTrunc: extended;
		var
			n: extended;
	begin
		GetLeftParen;
		n := GetExpression;
		GetRightParen;
		if Token <> DoneT then
			DoTrunc := trunc(n)
		else
			DoTrunc := 0.0;
	end;


	function DoRound: extended;
		var
			n: extended;
	begin
		GetLeftParen;
		n := GetExpression;
		GetRightParen;
		if Token <> DoneT then
			DoRound := round(n)
		else
			DoRound := 0.0;
	end;


	function GetNumber: extended; {(prompt:str255; default:extended)}
		var
			prompt: str255;
			default, n: extended;
	begin
		GetLeftParen;
		prompt := GetString;
		GetComma;
		default := GetExpression;
		GetRightParen;
		n := 0.0;
		if Token <> DoneT then begin
				n := GetReal(prompt, default);
				if n = BadReal then {cancel}
					token := DoneT;
			end;
		GetNumber := n;
	end;


	function GetInteger: LongInt;
		var
			n: LongInt;
			r: extended;
	begin
		r := GetExpression;
		if token = DoneT then begin
				GetInteger := 0;
				exit(GetInteger);
			end;
		GetInteger := trunc(r);
	end;


	function DoGetPixel: extended; {(hloc,vloc:integer)}
		var
			hloc, vloc: integer;
	begin
		GetLeftParen;
		hloc := GetInteger;
		GetComma;
		vloc := GetInteger;
		GetRightParen;
		if (Token <> DoneT) and (info <> NoInfo) then
			DoGetPixel := MyGetPixel(hloc, vloc)
		else
			DoGetPixel := 0.0;
	end;


	function ExecuteFunction: extended;
	begin
		case command of
			TruncC: 
				ExecuteFunction := DoTrunc;
			RoundC: 
				ExecuteFunction := DoRound;
			GetNumC: 
				ExecuteFunction := GetNumber;
			RandomC: 
				ExecuteFunction := (random + 32767.0) / 65534.0;
			GetPixelC: 
				ExecuteFunction := DoGetPixel;
			ButtonC: 
				ExecuteFunction := ord(Button);
		end; {case}
	end;


	function GetValue: extended;
	begin
		case token of
			Variable: 
				GetValue := TokenValue;
			NumericLiteral: 
				GetValue := StringToReal(TokenStr);
			FunctionT: 
				GetValue := ExecuteFunction;
			otherwise
				begin
					MacroError('Number expected');
					GetValue := 0.0;
					exit(GetValue);
				end;
		end; {case}
	end;


	function GetFactor: extended;
		var
			fValue: extended;
			isUnaryMinus, isNot: boolean;
	begin
		GetToken;
		isUnaryMinus := token = MinusOp;
		isNot := token = NotOp;
		if isUnaryMinus or isNot then
			GetToken;
		case token of
			Variable, NumericLiteral, FunctionT: 
				fValue := GetValue;
			LeftParen: 
				begin
					fValue := GetExpression;
					GetRightParen;
				end;
			otherwise
				begin
					macroError('Undefined identifier');
					fvalue := 0.0
				end;
		end;
		GetToken;
		if isUnaryMinus then
			fValue := -fValue;
		if isNot then
			if fValue = ord(true) then
				fValue := ord(false)
			else
				fValue := ord(true);
		GetFactor := fValue;
	end;


	function GetTerm: extended;
		var
			tValue, fValue: extended;
			op: TokenType;
	begin
		tValue := GetFactor;
		while token in [MulOp, IntDivOp, ModOp, DivOp, AndOp] do begin
				op := token;
				fValue := GetFactor;
				case op of
					MulOp: 
						tValue := tValue * fValue;
					IntDivOp: 
						if fValue <> 0.0 then
							tValue := trunc(tValue) div trunc(fValue)
						else
							MacroError(DivideByZero);
					ModOp: 
						if fValue <> 0.0 then
							tValue := trunc(tValue) mod trunc(fValue)
						else
							MacroError(DivideByZero);
					DivOp: 
						if fValue <> 0.0 then
							tValue := tValue / fValue
						else
							MacroError(DivideByZero);
					AndOp: 
						begin
							CheckBoolean(tValue);
							CheckBoolean(fValue);
							tValue := ord((tValue = ord(true)) and (fValue = ord(true)));
						end;
				end; {case}
			end; {while}
		GetTerm := tValue;
	end;


	function GetSimpleExpression: extended;
		var
			seValue, tValue: extended;
			op: TokenType;
	begin
		seValue := GetTerm;
		while token in [PlusOp, MinusOp, OrOp] do begin
				op := token;
				tValue := GetTerm;
				case op of
					PlusOp: 
						seValue := seValue + tValue;
					MinusOp: 
						seValue := seValue - tValue;
					orOp: 
						begin
							CheckBoolean(seValue);
							CheckBoolean(tValue);
							seValue := ord((seValue = ord(true)) or (tValue = ord(true)));
						end;
				end;
			end;
		GetSimpleExpression := seValue;
	end;


	function GetExpression: extended;
		var
			eValue, seValue: extended;
			op: TokenType;
	begin
		eValue := GetSimpleExpression;
		while token in [eqOp, ltOp, gtOp, neOp, leOp, geOp] do begin
				op := token;
				seValue := GetSimpleExpression;
				case op of
					eqOp: 
						eValue := ord(eValue = seValue);
					ltOp: 
						eValue := ord(eValue < seValue);
					gtOp: 
						eValue := ord(eValue > seValue);
					neOp: 
						eValue := ord(eValue <> seValue);
					leOp: 
						eValue := ord(eValue <= seValue);
					geOp: 
						eValue := ord(eValue >= seValue);
				end;
			end;
		GetExpression := eValue;
		PutTokenBack;
		;
	end;


	procedure DoWait;
		var
			seconds: extended;
			SaveTicks: LongInt;
			str: str255;
	begin
		GetLeftParen;
		seconds := GetExpression;
		GetRightParen;
		if Token <> DoneT then begin
				SaveTicks := TickCount + round(seconds * 60.0);
				repeat
				until (TickCount > SaveTicks) or CommandPeriod;
			end;
	end;


	procedure DoScaleSelection;
		var
			i: integer;
	begin
		GetLeftParen;
		hscale1 := GetExpression;
		GetComma;
		vscale1 := GetExpression;
		GetRightParen;
		if Token <> DoneT then begin
				ScaleSelection;
				KillRoi;
			end;
	end;


	procedure RangeCheck (i: integer);
	begin
		if (i < 0) or (i > 255) then
			MacroError('One or more arguments is less than 0 or greater than 255');
	end;


	procedure SetDensitySlice;
		var
			sStart, sEnd: integer;
	begin
		GetLeftParen;
		sStart := GetInteger;
		RangeCheck(sStart);
		GetComma;
		sEnd := GetInteger;
		RangeCheck(sEnd);
		if Token <> DoneT then begin
				SliceStart := sStart;
				SliceEnd := sEnd;
				if thresholding then
					ResetGrayMap;
				EnableDensitySlice;
			end;
		GetRightParen;
	end;


	procedure SetColor;
		var
			index: integer;
			SaveCommand: CommandType;
	begin
		SaveCommand := Command;
		GetLeftParen;
		index := GetInteger;
		GetRightParen;
		RangeCheck(index);
		if Token <> DoneT then begin
				if SaveCommand = SetForeC then
					SetForegroundColor(index)
				else
					SetBackgroundColor(index);
			end;
	end;


	procedure DoConstantArithmetic;
		var
			constant: integer;
	begin
		GetLeftParen;
		constant := GetInteger;
		DoArithmetic(AddItem, constant);
		GetRightParen;
	end;


	procedure GetNextWindow;
		var
			n: integer;
	begin
		n := info^.PicNum + 1;
		if n > nPics then
			n := 1;
		KillRoi;
		Info := pointer(WindowPeek(PicWindow[n])^.RefCon);
		SetPort(info^.wptr);
		IsInsertionPoint := false;
		WhatToUndo := NothingToUndo;
		UndoFromClip := false;
	end;


	procedure DoRevert;
	begin
		if revertable then begin
				RevertToSaved;
				UpdatePicWindow;
			end
		else
			MacroError('Unable to revert');
	end;


	procedure MakeRoi;
		var
			Left, Top, Width, Height: integer;
			SaveCommand: CommandType;
	begin
		SaveCommand := Command;
		GetLeftParen;
		left := GetInteger;
		GetComma;
		top := GetInteger;
		GetComma;
		width := GetInteger;
		GetComma;
		height := GetInteger;
		GetRightParen;
		KillRoi;
		if token <> DoneT then
			with Info^ do begin
					if SaveCommand = MakeOvalC then
						RoiType := OvalRoi
					else
						RoiType := RectRoi;
					SetRect(osroiRect, left, top, left + width, top + height);
					roiRect := osroiRect;
					OffscreenToScreenRect(roiRect);
					MakeRegion;
					SetupUndo;
					WhatToUndo := NothingToUndo;
					RoiShowing := true;
				end;
	end;


	procedure MoveRoi;
		var
			DeltaH, DeltaV: integer;
	begin
		GetLeftParen;
		DeltaH := GetInteger;
		GetComma;
		DeltaV := GetInteger;
		GetRightParen;
		with info^ do begin
				if not RoiShowing then begin
						MacroError('No Selection');
						exit(MoveRoi);
					end;
				OffsetRgn(osroiRgn, DeltaH, DeltaV);
				osroiRect := osroiRgn^^.rgnBBox;
				roiRect := osroiRect;
				OffscreenToScreenRect(roiRect);
				RoiUpdateTime := 0;
			end;
	end;


	procedure InsetRoi;
		var
			delta: integer;
	begin
		GetLeftParen;
		delta := GetInteger;
		GetRightParen;
		with info^ do begin
				if not RoiShowing then begin
						MacroError('No Selection');
						exit(InsetRoi);
					end;
				InsetRgn(osroiRgn, delta, delta);
				osroiRect := osroiRgn^^.rgnBBox;
				roiRect := osroiRect;
				OffscreenToScreenRect(roiRect);
				RoiUpdateTime := 0;
			end;
	end;


	procedure DoMoveTo;
	begin
		GetLeftParen;
		CurrentH := GetInteger;
		GetComma;
		CurrentV := GetInteger;
		GetRightParen;
	end;


	procedure Drawtext (str: str255);
	begin
		TextStr := str;
		InsertionPoint.h := CurrentH;
		InsertionPoint.v := CurrentV + 4;
		IsInsertionPoint := true;
		TextStart := InsertionPoint;
		KillRoi;
		DisplayText;
		IsInsertionPoint := false;
		CurrentV := CurrentV + CurrentSize;
	end;


	procedure DrawNumber;
		var
			n: extended;
			str: str255;
			fwidth: integer;
	begin
		GetLeftParen;
		n := GetExpression;
		GetRightParen;
		if token <> DoneT then begin
				if n = trunc(n) then
					fwidth := 0
				else
					fwidth := precision;
				RealToString(n, 1, fwidth, str);
				DrawText(str);
			end;
	end;


	procedure MakeLowerCase (var str: str255);
		var
			i: integer;
			c: char;
	begin
		for i := 1 to length(str) do begin
				c := str[i];
				if (c >= 'A') and (c <= 'Z') then
					str[i] := chr(ord(c) + 32);
			end;
	end;


	procedure SetFont;
		var
			FontName: str255;
			id: integer;
	begin
		GetLeftParen;
		FontName := GetString;
		GetRightParen;
		if Token <> DoneT then begin
				GetFNum(FontName, id);
				if id = 0 then
					MacroError('Font not available')
				else
					CurrentFontID := id;
				UpdateTextMenu
			end;
	end;


	procedure SetFontSize;
		var
			size: integer;
	begin
		GetLeftParen;
		Size := GetInteger;
		GetRightParen;
		if Token <> DoneT then begin
				CurrentSize := size;
				UpdateTextMenu
			end;
	end;


	procedure SetText;
		var
			Attributes: str255;
	begin
		GetLeftParen;
		Attributes := GetString;
		GetRightParen;
		if Token <> DoneT then begin
				MakeLowerCase(Attributes);
				if pos('with', Attributes) <> 0 then
					TextBack := WithBack;
				if pos('no', Attributes) <> 0 then
					TextBack := NoBack;
				if pos('left', Attributes) <> 0 then
					TextJust := teJustLeft;
				if pos('center', Attributes) <> 0 then
					TextJust := teJustCenter;
				if pos('right', Attributes) <> 0 then
					TextJust := teJustRight;
				CurrentStyle := [];
				if pos('bold', Attributes) <> 0 then
					CurrentStyle := CurrentStyle + [Bold];
				if pos('italic', Attributes) <> 0 then
					CurrentStyle := CurrentStyle + [Italic];
				if pos('underline', Attributes) <> 0 then
					CurrentStyle := CurrentStyle + [Underline];
				if pos('outline', Attributes) <> 0 then
					CurrentStyle := CurrentStyle + [Outline];
				if pos('shadow', Attributes) <> 0 then
					CurrentStyle := CurrentStyle + [Shadow];
				UpdateTextMenu
			end;
	end;


	procedure DoPutMessage;
		var
			str: str255;
	begin
		GetLeftParen;
		str := GetString;
		GetRightParen;
		if Token <> DoneT then
			PutMessage(str)
	end;


	function GetVar: integer;
	begin
		GetVar := 0;
		GetToken;
		if token <> Variable then
			MacroError('Variable expected')
		else
			GetVar := TokenStackLoc;
	end;


	procedure GetPicSize;  {(width,height)}
		var
			loc1, loc2: integer;
	begin
		GetLeftParen;
		loc1 := GetVar;
		GetComma;
		loc2 := GetVar;
		GetRightParen;
		if Token <> DoneT then
			with MacrosP^ do
				if info = NoInfo then begin
						stack[loc1].value := 0.0;
						stack[loc2].value := 0.0;
					end
				else
					with info^ do begin
							stack[loc1].value := PixelsPerLine;
							stack[loc2].value := nLines;
						end;
	end;


	procedure GetRoi;  {(hloc,vloc,width,height)}
		var
			loc1, loc2, loc3, loc4: integer;
	begin
		GetLeftParen;
		loc1 := GetVar;
		GetComma;
		loc2 := GetVar;
		GetComma;
		loc3 := GetVar;
		GetComma;
		loc4 := GetVar;
		GetRightParen;
		if Token <> DoneT then
			with MacrosP^, Info^ do
				if RoiShowing then
					with osroiRect do begin
							stack[loc1].value := left;
							stack[loc2].value := top;
							stack[loc3].value := right - left;
							stack[loc4].value := bottom - top;
						end
				else begin
						stack[loc1].value := 0.0;
						stack[loc2].value := 0.0;
						stack[loc3].value := 0.0;
						stack[loc4].value := 0.0;
					end;
	end;


	procedure CaptureOneFrame;
	begin
		if FrameGrabber <> QuickCapture then
			MacroError('DT2255 frame grabber not installed')
		else begin
				StartDigitizing;
				CaptureAndDisplayQCFrame;
				StopDigitizing;
			end;
	end;


	procedure DoMakeNewWindow; {(name:str255)}
		var
			name: str255;
	begin
		GetLeftParen;
		name := GetString;
		GetRightParen;
		if token <> DoneT then
			if not NewPicWindow(name, NewPicWidth, NewPicHeight) then
				MacroError('Out of memory');
	end;


	procedure SetPalette;
		var
			PaletteType: str255;
			ok: boolean;
	begin
		GetLeftParen;
		PaletteType := GetString;
		GetRightParen;
		if token <> DoneT then begin
				MakeLowerCase(PaletteType);
				if pos('gray', PaletteType) <> 0 then
					ResetGrayMap
				else if pos('pseudo', PaletteType) <> 0 then
					UpdateColors
				else if pos('system', PaletteType) <> 0 then
					ok := LoadCLUTResource(AppleDefaultCLUT)
				else if pos('spectrum', PaletteType) <> 0 then
					Load256ColorCLUT;
			end;
	end;


	procedure ExecuteCommand;
		var
			AutoSelectAll: boolean;
			t: FateTable;  {Needed for MakeSkeleton}
			str: str255;
	begin
		if (Info = NoInfo) and not (command in [GetPicSizeC, PutMsgC, ExitC, CaptureC, MoveToC, BeepC, MakeNewC, SetPaletteC, SetBackC]) then begin
				MacroError('No image window open');
				exit(ExecuteCommand);
			end;
		case command of
			RotateRC: 
				FlipOrRotate(RotateRight);
			RotateLC: 
				FlipOrRotate(RotateLeft);
			FlipVC: 
				FlipOrRotate(FlipVertical);
			FlipHC: 
				FlipOrRotate(FlipHorizontal);
			CopyC: 
				begin
					FindWhatToCopy;
					if WhatToCopy = NothingToCopy then
						MacroError('Copy failed')
					else
						DoCopy;
				end;
			SelectC: 
				SelectAll(true);
			ClearC, FillC, InvertC, FrameC, PasteC: 
				with info^ do begin
						AutoSelectAll := not RoiShowing;
						if AutoSelectAll then
							SelectAll(true);
						case command of
							ClearC: 
								DoOperation(EraseOp);
							FillC: 
								DoOperation(PaintOp);
							InvertC: 
								DoOperation(InvertOp);
							FrameC: 
								DoOperation(FrameOp);
							PasteC: 
								begin
									PastePicture;
									DoOperation(PasteOp);
								end;
						end;
						UpdateScreen(RoiRect);
					end;
			KillC: 
				KillRoi;
			RestoreC: 
				RestoreRoi;
			AnalyzeC: 
				AnalyzeParticles;
			ConvolveC: 
				DoConvolve;
			NextC: 
				GetNextWindow;
			MarkC: 
				MarkSelection(nRegions);
			MeasureC: 
				Measure;
			SliceC: 
				begin
					if thresholding then
						ResetGrayMap;
					EnableDensitySlice;
				end;
			NoSliceC: 
				DisableDensitySlice;
			MakeBinC: 
				MakeBinary;
			DitherC: 
				Filter(Dither, 0, t);
			SmoothC: 
				Filter(WeightedAvg, 0, t);
			SmoothMoreC: 
				Filter(UnweightedAvg, 0, t);
			SharpenC: 
				Filter(fsharpen, 0, t);
			SharpenMoreC: 
				begin
					OptionKeyWasDown := true;
					Filter(fsharpen, 0, t);
				end;
			ShadowC: 
				Filter(fshadow, 0, t);
			TraceC: 
				Filter(EdgeDetect, 0, t);
			DetectC: 
				begin
					OptionKeyWasDown := true;
					Filter(EdgeDetect, 0, t);
				end;
			RedirectC: 
				RedirectSampling := true;
			NoRedirectC: 
				RedirectSampling := false;
			ThresholdC: 
				EnableThresholding;
			ResetgmC: 
				ResetGrayMap;
			WaitC: 
				DoWait;
			ResetmC: 
				ResetCounters;
			SetSliceC: 
				SetDensitySlice;
			UndoC: 
				DoUndo;
			ScaleC: 
				DoScaleSelection;
			SetForeC, SetBackC: 
				SetColor;
			HistoC: 
				begin
					DoHistogram;
					DrawHistogram;
				end;
			EnhanceC: 
				EnhanceContrast;
			EqualizeC: 
				EqualizeHistogram;
			ErodeC: 
				DoErosion;
			DilateC: 
				DoDilation;
			OutlineC: 
				filter(OutlineFilter, 0, t);
			ThinC: 
				MakeSkeleton;
			AddConstantC: 
				DoConstantArithmetic;
			RevertC: 
				DoRevert;
			BeepC: 
				Beep;
			NopC: 
				;
			MakeC, MakeOvalC: 
				MakeRoi;
			MoveC: 
				MoveRoi;
			InsetC: 
				InsetRoi;
			MoveToC: 
				DoMoveTo;
			DrawTextC: 
				begin
					GetLeftParen;
					Str := GetString;
					GetRightParen;
					if token <> DoneT then
						DrawText(str);
				end;
			SetFontC: 
				SetFont;
			SetFontSizeC: 
				SetFontSize;
			SetTextC: 
				SetText;
			DrawNumC: 
				DrawNumber;
			ExitC: 
				token := DoneT;
			GetPicSizeC: 
				GetPicSize;
			PutMsgC: 
				DoPutMessage;
			GetRoiC: 
				GetRoi;
			CaptureC: 
				CaptureOneFrame;
			MakeNewC: 
				DoMakeNewWindow;
			DrawScaleC: 
				if info^.RoiShowing then begin
						DrawScale;
						UpdatePicWindow
					end
				else
					MacroError('No Selection');
			SetPaletteC: 
				SetPalette;
		end; {case}
		OptionKeyWasDown := false;
	end;


	procedure DoCompoundStatement;
	begin
		if token <> BeginT then
			MacroError('"begin" expected');
		GetToken;
		while (token <> endT) and (token <> DoneT) do begin
				DoStatement;
				GetToken;
				if Token = SemiColon then
					GetToken;
			end;
	end;


	procedure CheckForReservedWord;
		var
			str: str255;
	begin
		if (token = CommandT) or (token = FunctionT) then begin
				str := TokenSymbol;
				TrimString(str);
				MacroError(concat('"', str, '" is a reserved word'));
			end;
	end;


	procedure DoDeclaration;
		var
			SaveStackLoc, StackLoc: integer;
	begin
		SaveStackLoc := TopOfStack;
		while (token = UnknownIdentifier) or (token = Variable) or (token = comma) do begin
				TopOfStack := TopOfStack + 1;
				if TopOfStack > MaxStackSize then begin
						MacroError(StackOverflow);
						exit(DoDeclaration);
					end;
				with MacrosP^.stack[TopOfStack] do begin
						Symbol := TokenSymbol;
						value := 0.0;
					end;
				GetToken;
				if token = comma then
					GetToken;
			end; {while}
		CheckForReservedWord;
		if token <> colon then
			MacroError('":" expected');
		GetToken;
		if (token <> IntegerT) and (token <> RealT) then
			MacroError('"integer" or "real" expected');
		for StackLoc := SaveStackLoc + 1 to TopOfStack do
			with macrosP^.stack[StackLoc] do
				case token of
					IntegerT: 
						vType := IntVar;
					RealT: 
						vType := RealVar;
					otherwise
				end;
		GetToken;
		if Token = SemiColon then
			GetToken;
	end;


	procedure DoDeclarations;
	begin
		if token = SemiColon then
			GetToken;
		if token = VarT then begin
				GetToken;
				while ((token = UnknownIdentifier) or (token = Variable)) and (Token <> DoneT) do
					DoDeclaration;
				CheckForReservedWord;
			end;
	end;


	procedure DoFor;
		var
			SavePC, StackLoc: integer;
			StartValue, EndValue, i: LongInt;
	begin
		StackLoc := GetVar;
		GetToken;
		if token <> AssignOp then begin
				MacroError('":=" expected');
				exit(DoFor);
			end;
		StartValue := GetInteger;
		if token = DoneT then
			exit(DoFor);
		GetToken;
		if token <> ToT then begin
				MacroError('"to" expected');
				exit(DoFor);
			end;
		EndValue := GetInteger;
		if token = DoneT then
			exit(DoFor);
		GetToken;
		if token <> DoT then begin
				MacroError(DoExpected);
				exit(DoFor);
			end;
		SavePC := pc;
		for i := StartValue to EndValue do
			with MacrosP^ do begin
					Stack[StackLoc].value := i;
					pc := SavePC;
					GetToken;
					DoStatement;
					if Token = DoneT then
						leave;
				end;
	end;


	procedure DoAssignment;
		var
			SaveStackLoc: integer;
	begin
		SaveStackLoc := TokenStackLoc;
		GetToken;
		if token <> AssignOp then begin
				MacroError('":=" expected');
				exit(DoAssignment);
			end;
		MacrosP^.stack[SaveStackLoc].value := GetExpression;
	end;


	procedure SkipStatement (Statement: TokenType);
		var
			count: integer;
	begin
		GetToken;
		if token = beginT then begin
				count := 1;
				repeat
					GetToken;
					case token of
						beginT: 
							count := count + 1;
						endT: 
							count := count - 1;
						DoneT: 
							begin
								MacroError('"end" expected');
								exit(SkipStatement);
							end;
						otherwise
					end; {case}
				until count = 0;
			end
		else begin
				while (token <> SemiColon) and (token <> endT) and not ((Statement = IfT) and (token = ElseT)) do begin
						GetToken;
						if token = DoneT then begin
								MacroError('";"  or "end" expected');
								exit(SkipStatement);
							end;
					end; {while}
				PutTokenBack;
			end;
	end;


	function GetBoolean: boolean;
		var
			value: extended;
	begin
		value := GetExpression;
		CheckBoolean(value);
		GetBoolean := value = ord(true);
	end;


	procedure DoIf;
		var
			isTrue: boolean;
	begin
		isTrue := GetBoolean;
		GetToken;
		if token <> ThenT then
			MacroError('"then" expected');
		if isTrue then begin
				GetToken;
				DoStatement
			end
		else
			SkipStatement(IfT);
		GetToken;
		if token = elseT then begin
				if isTrue then
					SkipStatement(NullT)
				else begin
						GetToken;
						DoStatement;
					end;
			end
		else
			PutTokenBack;
	end;


	procedure DoWhile;
		var
			isTrue: boolean;
			SavePC: integer;
	begin
		SavePC := pc;
		repeat
			pc := SavePC;
			isTrue := GetBoolean;
			GetToken;
			if token <> doT then
				MacroError(DoExpected);
			if isTrue then begin
					GetToken;
					DoStatement
				end
			else
				SkipStatement(WhileT);
		until not isTrue or (Token = DoneT);
	end;


	procedure DoRepeat;
		var
			isTrue: boolean;
			SavePC: integer;
	begin
		SavePC := pc;
		isTrue := true;
		repeat
			pc := SavePC;
			GetToken;
			while (token <> untilT) and (token <> DoneT) do begin
					DoStatement;
					GetToken;
					if Token = SemiColon then
						GetToken;
				end;
			if token <> untilT then
				MacroError('"until" expected');
			isTrue := GetBoolean;
		until isTrue or (Token = DoneT);
	end;


	procedure PushArguments (var nArgs: integer);
		var
			arg: array[1..MaxArgs] of extended;
			i: integer;
			TempName: SymbolType;
	begin
		nArgs := 1;
		Arg[nArgs] := GetExpression;
		GetToken;
		while token = comma do begin
				if nArgs < MaxArgs then
					nArgs := nArgs + 1
				else
					MacroError('Too many arguments');
				arg[nArgs] := GetExpression;
				GetToken;
			end;
		if token <> RightParen then
			MacroError(RightParenExpected);
		for i := 1 to nArgs do begin
				if TopOfStack < MaxStackSize then
					TopOfStack := TopOfStack + 1
				else
					MacroError(StackOverflow);
				with MacrosP^.stack[TopOfStack] do begin
						vType := RealVar;
						value := arg[i];
					end;
			end;
	end;


	procedure DoProcedure;
		var
			SavePC, SavePCStart, SaveStackLoc, nArgs, i: integer;
			SaveProcName: SymbolType;
	begin
		SavePCStart := PCStart;
		PCStart := TokenLoc;
		SaveProcName := ProcName;
		ProcName := TokenSymbol;
		SaveStackLoc := TopOfStack;
		GetToken;
		if token = LeftParen then
			PushArguments(nArgs)
		else begin
				nArgs := 0;
				PutTokenBack;
			end;
		SavePC := pc;
		pc := pcStart;
		if nArgs > 0 then begin
				GetLeftParen;
				i := 0;
				GetToken;
				while token in [UnknownIdentifier, Variable, comma, colon, SemiColon, RealT, IntegerT] do begin
						if (token = UnknownIdentifier) or (token = Variable) then begin
								if i < nArgs then
									i := i + 1
								else
									MacroError('Too many arguments');
								MacrosP^.stack[SaveStackLoc + i].symbol := TokenSymbol;
							end;
						GetToken;
					end;
				if Token = VarT then
					MacroError('VAR parameters not supported');
				if i < nArgs then
					MacroError('Too few arguments');
				if token <> RightParen then
					MacroError(RightParenExpected);
			end;
		GetToken;
		if (token = LeftParen) and (nArgs = 0) then
			MacroError('Arguments not expected');
		DoDeclarations;
		DoCompoundStatement;
		pc := SavePC;
		TopOfStack := SaveStackLoc;
		pcStart := SavePCStart;
		ProcName := SaveProcName;
	end;


	procedure DoStatement;
	begin
		case token of
			BeginT: 
				DoCompoundStatement;
			CommandT: 
				ExecuteCommand;
			ForT: 
				DoFor;
			IfT: 
				DoIf;
			WhileT: 
				DoWhile;
			RepeatT: 
				DoRepeat;
			UnknownIdentifier: 
				MacroError('Unidentified identifier');
			Variable: 
				DoAssignment;
			ProcedureT: 
				DoProcedure;
			ElseT: 
				MacroError('Statement expected');
			otherwise
		end;
		if CommandPeriod then begin
				Token := DoneT;
				KillRoi;
				beep;
			end;
	end;


	procedure RunMacro (nMacro: integer);
		var
			count: integer;
			str: str255;
			SaveInfo: InfoPtr;
	begin
		StopDigitizing;
		str := '';
		count := 0;
		pcStart := MacroStart[nMacro];
		pc := pcStart;
		SavePC := pcStart;
		token := NullT;
		macro := true;
		SaveInfo := info;
		TopOfStack := 0;
		ProcName := '        ';
		GetToken;
		DoDeclarations;
		DoCompoundStatement;
		if (info <> SaveInfo) and (info <> NoInfo) then
			SelectWindow(info^.wptr);
		with info^, osroiRect do begin
				if ((right - left) <= 0) or ((bottom - top) <= 0) then
					KillRoi;
			end;
		if info^.RoiShowing then
			UpdatePicWindow;
		macro := false;
		exit(RunMacro);
		str := '';
		while (pc <= EndMacros) and (count <= 100) do begin
				str := concat(str, MacrosP^.macros[pc]);
				pc := pc + 1;
				count := count + 1;
			end;
		ShowMessage(str);
	end;


	procedure RunKeyMacro (ch: char; KeyCode: integer);
		const
			FunctionKey = 16;
		var
			i: integer;
	begin
		if (ord(ch) = 0) then
			exit(RunKeyMacro);
		if ord(ch) = FunctionKey then
			case KeyCode of
				122: 
					ch := 'A';
				120: 
					ch := 'B';
				99: 
					ch := 'C';
				118: 
					ch := 'D';
				96: 
					ch := 'E';
				97: 
					ch := 'F';
				98: 
					ch := 'G';
				100: 
					ch := 'H';
				101: 
					ch := 'I';
				109: 
					ch := 'J';
				103: 
					ch := 'K';
				111: 
					ch := 'L';
				105: 
					ch := 'M';
				107: 
					ch := 'N';
				113: 
					ch := 'O';
				otherwise
			end;
		for i := 1 to nMacros do
			if ch = MacroKey[i] then begin
					RunMacro(i);
					leave;
				end;
	end;


end.